// $Id: FormatReader.java 3594 2008-03-28 14:55:34Z yura $

package org.refal.plus.compiler;

import java.io.FileInputStream;
import java.io.InputStream;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

public class FormatReader {
    public static String ReadFormatsFromClass (InputStream is) {
        try {
            return (new FormatReader()).ReadFormatsFromClass(new ClassReader(is));
        } catch (Exception exc) {
            return null;
        }
    }

    public static String ReadFormatsFromClass (String classname) {
        try {
            InputStream inputStream = FormatReader.class.getClassLoader().getResourceAsStream(classname + ".class");
            String res = ReadFormatsFromClass(inputStream);
            inputStream.close();
            return res;
        } catch (Exception exc) {
            return null;
        }
    }

    public static String ReadFormatsFromFile (String filename) {
        try {
            FileInputStream fileInputStream = new FileInputStream(filename);
            String res = ReadFormatsFromClass(fileInputStream);
            fileInputStream.close();
            return res;
        } catch (Exception exc) {
            return null;
        }
    }

    public static void main (String[] args) throws Exception {
        for (int i = 0; i < args.length; i++) {
            System.out.println(args[i]);
            System.out.println(ReadFormatsFromClass(args[i]));
        }
    }

    private class MethodFormatAnnotationReader implements AnnotationVisitor {
        public MethodFormatAnnotationReader () {}

        public void visit (String name, Object value) {
            if (name.equals("value") && value instanceof String) {
                if (!formats.equals(""))
                    formats += " ";
                formats += ((String) value);
            }
        }

        public AnnotationVisitor visitAnnotation (String name, String desc) {
            return null;
        }

        public AnnotationVisitor visitArray (String name) {
            return null;
        }

        public void visitEnd () {}

        public void visitEnum (String name, String desc, String value) {}
    }

    private class FieldFormatReader implements FieldVisitor {
        public FieldFormatReader () {}

        public AnnotationVisitor visitAnnotation (String desc, boolean visible) {
            if (desc.equals("Lorg/refal/plus/RefalFormat;"))
                return new MethodFormatAnnotationReader();
            else
                return null;
        }

        public void visitAttribute (Attribute attr) {}

        public void visitEnd () {}
    }

    private class MethodFormatReader implements MethodVisitor {
        public MethodFormatReader () {}

        public AnnotationVisitor visitAnnotation (String desc, boolean visible) {
            if (desc.equals("Lorg/refal/plus/RefalFormat;"))
                return new MethodFormatAnnotationReader();
            else
                return null;
        }

        public AnnotationVisitor visitAnnotationDefault () {
            return null;
        }

        public void visitAttribute (Attribute attr) {}

        public void visitCode () {}

        public void visitEnd () {}

        public void visitFieldInsn (int opcode, String owner, String name, String desc) {}

        public void visitFrame (int type, int nLocal, Object[] local, int nStack, Object[] stack) {}

        public void visitIincInsn (int var, int increment) {}

        public void visitInsn (int opcode) {}

        public void visitIntInsn (int opcode, int operand) {}

        public void visitJumpInsn (int opcode, Label label) {}

        public void visitLabel (Label label) {}

        public void visitLdcInsn (Object cst) {}

        public void visitLineNumber (int line, Label start) {}

        public void visitLocalVariable (String name, String desc, String signature, Label start, Label end, int index) {}

        public void visitLookupSwitchInsn (Label dflt, int[] keys, Label[] labels) {}

        public void visitMaxs (int maxStack, int maxLocals) {}

        public void visitMethodInsn (int opcode, String owner, String name, String desc) {}

        public void visitMultiANewArrayInsn (String desc, int dims) {}

        public AnnotationVisitor visitParameterAnnotation (int parameter, String desc, boolean visible) {
            return null;
        }

        public void visitTableSwitchInsn (int min, int max, Label dflt, Label[] labels) {}

        public void visitTryCatchBlock (Label start, Label end, Label handler, String type) {}

        public void visitTypeInsn (int opcode, String desc) {}

        public void visitVarInsn (int opcode, int var) {}
    }

    private class ClassFormatReader implements ClassVisitor {
        public ClassFormatReader () {}

        public void visit (int version, int access, String name, String signature, String superName, String[] interfaces) {
            if (superName != null) {
                String fmts = ReadFormatsFromClass(superName);
                if (fmts != null) {
                    if (!formats.equals(""))
                        formats += " ";
                    formats += fmts;
                }
            }
        }

        public AnnotationVisitor visitAnnotation (String desc, boolean visible) {
            return null;
        }

        public void visitAttribute (Attribute attr) {}

        public void visitEnd () {}

        public FieldVisitor visitField (int access, String name, String desc, String signature, Object value) {
            return ((access & Opcodes.ACC_PUBLIC) != 0) ? new FieldFormatReader() : null;
        }

        public void visitInnerClass (String name, String outerName, String innerName, int access) {}

        public MethodVisitor visitMethod (int access, String name, String desc, String signature, String[] exceptions) {
            return ((access & Opcodes.ACC_PUBLIC) != 0) ? new MethodFormatReader() : null;
        }

        public void visitOuterClass (String owner, String name, String desc) {}

        public void visitSource (String source, String debug) {}
    }
    
    private String formats;
    
    private FormatReader () {
        formats = "";
    }

    private String ReadFormatsFromClass (ClassReader cr) {
        ClassFormatReader cv = new ClassFormatReader();
        cr.accept(cv, 0);
        return formats;
    }
}
